package com.termux.terminal

/**
 * Implementation of wcwidth(3) for Unicode 15.
 *
 * Implementation from https://github.com/jquast/wcwidth but we return 0 for unprintable characters.
 *
 * IMPORTANT:
 * Must be kept in sync with the following:
 * https://github.com/termux/wcwidth
 * https://github.com/termux/libandroid-support
 * https://github.com/termux/termux-packages/tree/master/packages/libandroid-support
 */
object WcWidth {
    // From https://github.com/jquast/wcwidth/blob/master/wcwidth/table_zero.py
    // from https://github.com/jquast/wcwidth/pull/64
    // at commit 1b9b6585b0080ea5cb88dc9815796505724793fe (2022-12-16):
    private val ZERO_WIDTH = arrayOf(
        intArrayOf(0x00300, 0x0036f),
        intArrayOf(0x00483, 0x00489),
        intArrayOf(0x00591, 0x005bd),
        intArrayOf(0x005bf, 0x005bf),
        intArrayOf(0x005c1, 0x005c2),
        intArrayOf(0x005c4, 0x005c5),
        intArrayOf(0x005c7, 0x005c7),
        intArrayOf(0x00610, 0x0061a),
        intArrayOf(0x0064b, 0x0065f),
        intArrayOf(0x00670, 0x00670),
        intArrayOf(0x006d6, 0x006dc),
        intArrayOf(0x006df, 0x006e4),
        intArrayOf(0x006e7, 0x006e8),
        intArrayOf(0x006ea, 0x006ed),
        intArrayOf(0x00711, 0x00711),
        intArrayOf(0x00730, 0x0074a),
        intArrayOf(0x007a6, 0x007b0),
        intArrayOf(0x007eb, 0x007f3),
        intArrayOf(0x007fd, 0x007fd),
        intArrayOf(0x00816, 0x00819),
        intArrayOf(0x0081b, 0x00823),
        intArrayOf(0x00825, 0x00827),
        intArrayOf(0x00829, 0x0082d),
        intArrayOf(0x00859, 0x0085b),
        intArrayOf(0x00898, 0x0089f),
        intArrayOf(0x008ca, 0x008e1),
        intArrayOf(0x008e3, 0x00902),
        intArrayOf(0x0093a, 0x0093a),
        intArrayOf(0x0093c, 0x0093c),
        intArrayOf(0x00941, 0x00948),
        intArrayOf(0x0094d, 0x0094d),
        intArrayOf(0x00951, 0x00957),
        intArrayOf(0x00962, 0x00963),
        intArrayOf(0x00981, 0x00981),
        intArrayOf(0x009bc, 0x009bc),
        intArrayOf(0x009c1, 0x009c4),
        intArrayOf(0x009cd, 0x009cd),
        intArrayOf(0x009e2, 0x009e3),
        intArrayOf(0x009fe, 0x009fe),
        intArrayOf(0x00a01, 0x00a02),
        intArrayOf(0x00a3c, 0x00a3c),
        intArrayOf(0x00a41, 0x00a42),
        intArrayOf(0x00a47, 0x00a48),
        intArrayOf(0x00a4b, 0x00a4d),
        intArrayOf(0x00a51, 0x00a51),
        intArrayOf(0x00a70, 0x00a71),
        intArrayOf(0x00a75, 0x00a75),
        intArrayOf(0x00a81, 0x00a82),
        intArrayOf(0x00abc, 0x00abc),
        intArrayOf(0x00ac1, 0x00ac5),
        intArrayOf(0x00ac7, 0x00ac8),
        intArrayOf(0x00acd, 0x00acd),
        intArrayOf(0x00ae2, 0x00ae3),
        intArrayOf(0x00afa, 0x00aff),
        intArrayOf(0x00b01, 0x00b01),
        intArrayOf(0x00b3c, 0x00b3c),
        intArrayOf(0x00b3f, 0x00b3f),
        intArrayOf(0x00b41, 0x00b44),
        intArrayOf(0x00b4d, 0x00b4d),
        intArrayOf(0x00b55, 0x00b56),
        intArrayOf(0x00b62, 0x00b63),
        intArrayOf(0x00b82, 0x00b82),
        intArrayOf(0x00bc0, 0x00bc0),
        intArrayOf(0x00bcd, 0x00bcd),
        intArrayOf(0x00c00, 0x00c00),
        intArrayOf(0x00c04, 0x00c04),
        intArrayOf(0x00c3c, 0x00c3c),
        intArrayOf(0x00c3e, 0x00c40),
        intArrayOf(0x00c46, 0x00c48),
        intArrayOf(0x00c4a, 0x00c4d),
        intArrayOf(0x00c55, 0x00c56),
        intArrayOf(0x00c62, 0x00c63),
        intArrayOf(0x00c81, 0x00c81),
        intArrayOf(0x00cbc, 0x00cbc),
        intArrayOf(0x00cbf, 0x00cbf),
        intArrayOf(0x00cc6, 0x00cc6),
        intArrayOf(0x00ccc, 0x00ccd),
        intArrayOf(0x00ce2, 0x00ce3),
        intArrayOf(0x00d00, 0x00d01),
        intArrayOf(0x00d3b, 0x00d3c),
        intArrayOf(0x00d41, 0x00d44),
        intArrayOf(0x00d4d, 0x00d4d),
        intArrayOf(0x00d62, 0x00d63),
        intArrayOf(0x00d81, 0x00d81),
        intArrayOf(0x00dca, 0x00dca),
        intArrayOf(0x00dd2, 0x00dd4),
        intArrayOf(0x00dd6, 0x00dd6),
        intArrayOf(0x00e31, 0x00e31),
        intArrayOf(0x00e34, 0x00e3a),
        intArrayOf(0x00e47, 0x00e4e),
        intArrayOf(0x00eb1, 0x00eb1),
        intArrayOf(0x00eb4, 0x00ebc),
        intArrayOf(0x00ec8, 0x00ece),
        intArrayOf(0x00f18, 0x00f19),
        intArrayOf(0x00f35, 0x00f35),
        intArrayOf(0x00f37, 0x00f37),
        intArrayOf(0x00f39, 0x00f39),
        intArrayOf(0x00f71, 0x00f7e),
        intArrayOf(0x00f80, 0x00f84),
        intArrayOf(0x00f86, 0x00f87),
        intArrayOf(0x00f8d, 0x00f97),
        intArrayOf(0x00f99, 0x00fbc),
        intArrayOf(0x00fc6, 0x00fc6),
        intArrayOf(0x0102d, 0x01030),
        intArrayOf(0x01032, 0x01037),
        intArrayOf(0x01039, 0x0103a),
        intArrayOf(0x0103d, 0x0103e),
        intArrayOf(0x01058, 0x01059),
        intArrayOf(0x0105e, 0x01060),
        intArrayOf(0x01071, 0x01074),
        intArrayOf(0x01082, 0x01082),
        intArrayOf(0x01085, 0x01086),
        intArrayOf(0x0108d, 0x0108d),
        intArrayOf(0x0109d, 0x0109d),
        intArrayOf(0x0135d, 0x0135f),
        intArrayOf(0x01712, 0x01714),
        intArrayOf(0x01732, 0x01733),
        intArrayOf(0x01752, 0x01753),
        intArrayOf(0x01772, 0x01773),
        intArrayOf(0x017b4, 0x017b5),
        intArrayOf(0x017b7, 0x017bd),
        intArrayOf(0x017c6, 0x017c6),
        intArrayOf(0x017c9, 0x017d3),
        intArrayOf(0x017dd, 0x017dd),
        intArrayOf(0x0180b, 0x0180d),
        intArrayOf(0x0180f, 0x0180f),
        intArrayOf(0x01885, 0x01886),
        intArrayOf(0x018a9, 0x018a9),
        intArrayOf(0x01920, 0x01922),
        intArrayOf(0x01927, 0x01928),
        intArrayOf(0x01932, 0x01932),
        intArrayOf(0x01939, 0x0193b),
        intArrayOf(0x01a17, 0x01a18),
        intArrayOf(0x01a1b, 0x01a1b),
        intArrayOf(0x01a56, 0x01a56),
        intArrayOf(0x01a58, 0x01a5e),
        intArrayOf(0x01a60, 0x01a60),
        intArrayOf(0x01a62, 0x01a62),
        intArrayOf(0x01a65, 0x01a6c),
        intArrayOf(0x01a73, 0x01a7c),
        intArrayOf(0x01a7f, 0x01a7f),
        intArrayOf(0x01ab0, 0x01ace),
        intArrayOf(0x01b00, 0x01b03),
        intArrayOf(0x01b34, 0x01b34),
        intArrayOf(0x01b36, 0x01b3a),
        intArrayOf(0x01b3c, 0x01b3c),
        intArrayOf(0x01b42, 0x01b42),
        intArrayOf(0x01b6b, 0x01b73),
        intArrayOf(0x01b80, 0x01b81),
        intArrayOf(0x01ba2, 0x01ba5),
        intArrayOf(0x01ba8, 0x01ba9),
        intArrayOf(0x01bab, 0x01bad),
        intArrayOf(0x01be6, 0x01be6),
        intArrayOf(0x01be8, 0x01be9),
        intArrayOf(0x01bed, 0x01bed),
        intArrayOf(0x01bef, 0x01bf1),
        intArrayOf(0x01c2c, 0x01c33),
        intArrayOf(0x01c36, 0x01c37),
        intArrayOf(0x01cd0, 0x01cd2),
        intArrayOf(0x01cd4, 0x01ce0),
        intArrayOf(0x01ce2, 0x01ce8),
        intArrayOf(0x01ced, 0x01ced),
        intArrayOf(0x01cf4, 0x01cf4),
        intArrayOf(0x01cf8, 0x01cf9),
        intArrayOf(0x01dc0, 0x01dff),
        intArrayOf(0x020d0, 0x020f0),
        intArrayOf(0x02cef, 0x02cf1),
        intArrayOf(0x02d7f, 0x02d7f),
        intArrayOf(0x02de0, 0x02dff),
        intArrayOf(0x0302a, 0x0302d),
        intArrayOf(0x03099, 0x0309a),
        intArrayOf(0x0a66f, 0x0a672),
        intArrayOf(0x0a674, 0x0a67d),
        intArrayOf(0x0a69e, 0x0a69f),
        intArrayOf(0x0a6f0, 0x0a6f1),
        intArrayOf(0x0a802, 0x0a802),
        intArrayOf(0x0a806, 0x0a806),
        intArrayOf(0x0a80b, 0x0a80b),
        intArrayOf(0x0a825, 0x0a826),
        intArrayOf(0x0a82c, 0x0a82c),
        intArrayOf(0x0a8c4, 0x0a8c5),
        intArrayOf(0x0a8e0, 0x0a8f1),
        intArrayOf(0x0a8ff, 0x0a8ff),
        intArrayOf(0x0a926, 0x0a92d),
        intArrayOf(0x0a947, 0x0a951),
        intArrayOf(0x0a980, 0x0a982),
        intArrayOf(0x0a9b3, 0x0a9b3),
        intArrayOf(0x0a9b6, 0x0a9b9),
        intArrayOf(0x0a9bc, 0x0a9bd),
        intArrayOf(0x0a9e5, 0x0a9e5),
        intArrayOf(0x0aa29, 0x0aa2e),
        intArrayOf(0x0aa31, 0x0aa32),
        intArrayOf(0x0aa35, 0x0aa36),
        intArrayOf(0x0aa43, 0x0aa43),
        intArrayOf(0x0aa4c, 0x0aa4c),
        intArrayOf(0x0aa7c, 0x0aa7c),
        intArrayOf(0x0aab0, 0x0aab0),
        intArrayOf(0x0aab2, 0x0aab4),
        intArrayOf(0x0aab7, 0x0aab8),
        intArrayOf(0x0aabe, 0x0aabf),
        intArrayOf(0x0aac1, 0x0aac1),
        intArrayOf(0x0aaec, 0x0aaed),
        intArrayOf(0x0aaf6, 0x0aaf6),
        intArrayOf(0x0abe5, 0x0abe5),
        intArrayOf(0x0abe8, 0x0abe8),
        intArrayOf(0x0abed, 0x0abed),
        intArrayOf(0x0fb1e, 0x0fb1e),
        intArrayOf(0x0fe00, 0x0fe0f),
        intArrayOf(0x0fe20, 0x0fe2f),
        intArrayOf(0x101fd, 0x101fd),
        intArrayOf(0x102e0, 0x102e0),
        intArrayOf(0x10376, 0x1037a),
        intArrayOf(0x10a01, 0x10a03),
        intArrayOf(0x10a05, 0x10a06),
        intArrayOf(0x10a0c, 0x10a0f),
        intArrayOf(0x10a38, 0x10a3a),
        intArrayOf(0x10a3f, 0x10a3f),
        intArrayOf(0x10ae5, 0x10ae6),
        intArrayOf(0x10d24, 0x10d27),
        intArrayOf(0x10eab, 0x10eac),
        intArrayOf(0x10efd, 0x10eff),
        intArrayOf(0x10f46, 0x10f50),
        intArrayOf(0x10f82, 0x10f85),
        intArrayOf(0x11001, 0x11001),
        intArrayOf(0x11038, 0x11046),
        intArrayOf(0x11070, 0x11070),
        intArrayOf(0x11073, 0x11074),
        intArrayOf(0x1107f, 0x11081),
        intArrayOf(0x110b3, 0x110b6),
        intArrayOf(0x110b9, 0x110ba),
        intArrayOf(0x110c2, 0x110c2),
        intArrayOf(0x11100, 0x11102),
        intArrayOf(0x11127, 0x1112b),
        intArrayOf(0x1112d, 0x11134),
        intArrayOf(0x11173, 0x11173),
        intArrayOf(0x11180, 0x11181),
        intArrayOf(0x111b6, 0x111be),
        intArrayOf(0x111c9, 0x111cc),
        intArrayOf(0x111cf, 0x111cf),
        intArrayOf(0x1122f, 0x11231),
        intArrayOf(0x11234, 0x11234),
        intArrayOf(0x11236, 0x11237),
        intArrayOf(0x1123e, 0x1123e),
        intArrayOf(0x11241, 0x11241),
        intArrayOf(0x112df, 0x112df),
        intArrayOf(0x112e3, 0x112ea),
        intArrayOf(0x11300, 0x11301),
        intArrayOf(0x1133b, 0x1133c),
        intArrayOf(0x11340, 0x11340),
        intArrayOf(0x11366, 0x1136c),
        intArrayOf(0x11370, 0x11374),
        intArrayOf(0x11438, 0x1143f),
        intArrayOf(0x11442, 0x11444),
        intArrayOf(0x11446, 0x11446),
        intArrayOf(0x1145e, 0x1145e),
        intArrayOf(0x114b3, 0x114b8),
        intArrayOf(0x114ba, 0x114ba),
        intArrayOf(0x114bf, 0x114c0),
        intArrayOf(0x114c2, 0x114c3),
        intArrayOf(0x115b2, 0x115b5),
        intArrayOf(0x115bc, 0x115bd),
        intArrayOf(0x115bf, 0x115c0),
        intArrayOf(0x115dc, 0x115dd),
        intArrayOf(0x11633, 0x1163a),
        intArrayOf(0x1163d, 0x1163d),
        intArrayOf(0x1163f, 0x11640),
        intArrayOf(0x116ab, 0x116ab),
        intArrayOf(0x116ad, 0x116ad),
        intArrayOf(0x116b0, 0x116b5),
        intArrayOf(0x116b7, 0x116b7),
        intArrayOf(0x1171d, 0x1171f),
        intArrayOf(0x11722, 0x11725),
        intArrayOf(0x11727, 0x1172b),
        intArrayOf(0x1182f, 0x11837),
        intArrayOf(0x11839, 0x1183a),
        intArrayOf(0x1193b, 0x1193c),
        intArrayOf(0x1193e, 0x1193e),
        intArrayOf(0x11943, 0x11943),
        intArrayOf(0x119d4, 0x119d7),
        intArrayOf(0x119da, 0x119db),
        intArrayOf(0x119e0, 0x119e0),
        intArrayOf(0x11a01, 0x11a0a),
        intArrayOf(0x11a33, 0x11a38),
        intArrayOf(0x11a3b, 0x11a3e),
        intArrayOf(0x11a47, 0x11a47),
        intArrayOf(0x11a51, 0x11a56),
        intArrayOf(0x11a59, 0x11a5b),
        intArrayOf(0x11a8a, 0x11a96),
        intArrayOf(0x11a98, 0x11a99),
        intArrayOf(0x11c30, 0x11c36),
        intArrayOf(0x11c38, 0x11c3d),
        intArrayOf(0x11c3f, 0x11c3f),
        intArrayOf(0x11c92, 0x11ca7),
        intArrayOf(0x11caa, 0x11cb0),
        intArrayOf(0x11cb2, 0x11cb3),
        intArrayOf(0x11cb5, 0x11cb6),
        intArrayOf(0x11d31, 0x11d36),
        intArrayOf(0x11d3a, 0x11d3a),
        intArrayOf(0x11d3c, 0x11d3d),
        intArrayOf(0x11d3f, 0x11d45),
        intArrayOf(0x11d47, 0x11d47),
        intArrayOf(0x11d90, 0x11d91),
        intArrayOf(0x11d95, 0x11d95),
        intArrayOf(0x11d97, 0x11d97),
        intArrayOf(0x11ef3, 0x11ef4),
        intArrayOf(0x11f00, 0x11f01),
        intArrayOf(0x11f36, 0x11f3a),
        intArrayOf(0x11f40, 0x11f40),
        intArrayOf(0x11f42, 0x11f42),
        intArrayOf(0x13440, 0x13440),
        intArrayOf(0x13447, 0x13455),
        intArrayOf(0x16af0, 0x16af4),
        intArrayOf(0x16b30, 0x16b36),
        intArrayOf(0x16f4f, 0x16f4f),
        intArrayOf(0x16f8f, 0x16f92),
        intArrayOf(0x16fe4, 0x16fe4),
        intArrayOf(0x1bc9d, 0x1bc9e),
        intArrayOf(0x1cf00, 0x1cf2d),
        intArrayOf(0x1cf30, 0x1cf46),
        intArrayOf(0x1d167, 0x1d169),
        intArrayOf(0x1d17b, 0x1d182),
        intArrayOf(0x1d185, 0x1d18b),
        intArrayOf(0x1d1aa, 0x1d1ad),
        intArrayOf(0x1d242, 0x1d244),
        intArrayOf(0x1da00, 0x1da36),
        intArrayOf(0x1da3b, 0x1da6c),
        intArrayOf(0x1da75, 0x1da75),
        intArrayOf(0x1da84, 0x1da84),
        intArrayOf(0x1da9b, 0x1da9f),
        intArrayOf(0x1daa1, 0x1daaf),
        intArrayOf(0x1e000, 0x1e006),
        intArrayOf(0x1e008, 0x1e018),
        intArrayOf(0x1e01b, 0x1e021),
        intArrayOf(0x1e023, 0x1e024),
        intArrayOf(0x1e026, 0x1e02a),
        intArrayOf(0x1e08f, 0x1e08f),
        intArrayOf(0x1e130, 0x1e136),
        intArrayOf(0x1e2ae, 0x1e2ae),
        intArrayOf(0x1e2ec, 0x1e2ef),
        intArrayOf(0x1e4ec, 0x1e4ef),
        intArrayOf(0x1e8d0, 0x1e8d6),
        intArrayOf(0x1e944, 0x1e94a),
        intArrayOf(0xe0100, 0xe01ef)
    )

    // https://github.com/jquast/wcwidth/blob/master/wcwidth/table_wide.py
    // from https://github.com/jquast/wcwidth/pull/64
    // at commit 1b9b6585b0080ea5cb88dc9815796505724793fe (2022-12-16):
    private val WIDE_EASTASIAN = arrayOf(
        intArrayOf(0x01100, 0x0115f),
        intArrayOf(0x0231a, 0x0231b),
        intArrayOf(0x02329, 0x0232a),
        intArrayOf(0x023e9, 0x023ec),
        intArrayOf(0x023f0, 0x023f0),
        intArrayOf(0x023f3, 0x023f3),
        intArrayOf(0x025fd, 0x025fe),
        intArrayOf(0x02614, 0x02615),
        intArrayOf(0x02648, 0x02653),
        intArrayOf(0x0267f, 0x0267f),
        intArrayOf(0x02693, 0x02693),
        intArrayOf(0x026a1, 0x026a1),
        intArrayOf(0x026aa, 0x026ab),
        intArrayOf(0x026bd, 0x026be),
        intArrayOf(0x026c4, 0x026c5),
        intArrayOf(0x026ce, 0x026ce),
        intArrayOf(0x026d4, 0x026d4),
        intArrayOf(0x026ea, 0x026ea),
        intArrayOf(0x026f2, 0x026f3),
        intArrayOf(0x026f5, 0x026f5),
        intArrayOf(0x026fa, 0x026fa),
        intArrayOf(0x026fd, 0x026fd),
        intArrayOf(0x02705, 0x02705),
        intArrayOf(0x0270a, 0x0270b),
        intArrayOf(0x02728, 0x02728),
        intArrayOf(0x0274c, 0x0274c),
        intArrayOf(0x0274e, 0x0274e),
        intArrayOf(0x02753, 0x02755),
        intArrayOf(0x02757, 0x02757),
        intArrayOf(0x02795, 0x02797),
        intArrayOf(0x027b0, 0x027b0),
        intArrayOf(0x027bf, 0x027bf),
        intArrayOf(0x02b1b, 0x02b1c),
        intArrayOf(0x02b50, 0x02b50),
        intArrayOf(0x02b55, 0x02b55),
        intArrayOf(0x02e80, 0x02e99),
        intArrayOf(0x02e9b, 0x02ef3),
        intArrayOf(0x02f00, 0x02fd5),
        intArrayOf(0x02ff0, 0x02ffb),
        intArrayOf(0x03000, 0x0303e),
        intArrayOf(0x03041, 0x03096),
        intArrayOf(0x03099, 0x030ff),
        intArrayOf(0x03105, 0x0312f),
        intArrayOf(0x03131, 0x0318e),
        intArrayOf(0x03190, 0x031e3),
        intArrayOf(0x031f0, 0x0321e),
        intArrayOf(0x03220, 0x03247),
        intArrayOf(0x03250, 0x04dbf),
        intArrayOf(0x04e00, 0x0a48c),
        intArrayOf(0x0a490, 0x0a4c6),
        intArrayOf(0x0a960, 0x0a97c),
        intArrayOf(0x0ac00, 0x0d7a3),
        intArrayOf(0x0f900, 0x0faff),
        intArrayOf(0x0fe10, 0x0fe19),
        intArrayOf(0x0fe30, 0x0fe52),
        intArrayOf(0x0fe54, 0x0fe66),
        intArrayOf(0x0fe68, 0x0fe6b),
        intArrayOf(0x0ff01, 0x0ff60),
        intArrayOf(0x0ffe0, 0x0ffe6),
        intArrayOf(0x16fe0, 0x16fe4),
        intArrayOf(0x16ff0, 0x16ff1),
        intArrayOf(0x17000, 0x187f7),
        intArrayOf(0x18800, 0x18cd5),
        intArrayOf(0x18d00, 0x18d08),
        intArrayOf(0x1aff0, 0x1aff3),
        intArrayOf(0x1aff5, 0x1affb),
        intArrayOf(0x1affd, 0x1affe),
        intArrayOf(0x1b000, 0x1b122),
        intArrayOf(0x1b132, 0x1b132),
        intArrayOf(0x1b150, 0x1b152),
        intArrayOf(0x1b155, 0x1b155),
        intArrayOf(0x1b164, 0x1b167),
        intArrayOf(0x1b170, 0x1b2fb),
        intArrayOf(0x1f004, 0x1f004),
        intArrayOf(0x1f0cf, 0x1f0cf),
        intArrayOf(0x1f18e, 0x1f18e),
        intArrayOf(0x1f191, 0x1f19a),
        intArrayOf(0x1f200, 0x1f202),
        intArrayOf(0x1f210, 0x1f23b),
        intArrayOf(0x1f240, 0x1f248),
        intArrayOf(0x1f250, 0x1f251),
        intArrayOf(0x1f260, 0x1f265),
        intArrayOf(0x1f300, 0x1f320),
        intArrayOf(0x1f32d, 0x1f335),
        intArrayOf(0x1f337, 0x1f37c),
        intArrayOf(0x1f37e, 0x1f393),
        intArrayOf(0x1f3a0, 0x1f3ca),
        intArrayOf(0x1f3cf, 0x1f3d3),
        intArrayOf(0x1f3e0, 0x1f3f0),
        intArrayOf(0x1f3f4, 0x1f3f4),
        intArrayOf(0x1f3f8, 0x1f43e),
        intArrayOf(0x1f440, 0x1f440),
        intArrayOf(0x1f442, 0x1f4fc),
        intArrayOf(0x1f4ff, 0x1f53d),
        intArrayOf(0x1f54b, 0x1f54e),
        intArrayOf(0x1f550, 0x1f567),
        intArrayOf(0x1f57a, 0x1f57a),
        intArrayOf(0x1f595, 0x1f596),
        intArrayOf(0x1f5a4, 0x1f5a4),
        intArrayOf(0x1f5fb, 0x1f64f),
        intArrayOf(0x1f680, 0x1f6c5),
        intArrayOf(0x1f6cc, 0x1f6cc),
        intArrayOf(0x1f6d0, 0x1f6d2),
        intArrayOf(0x1f6d5, 0x1f6d7),
        intArrayOf(0x1f6dc, 0x1f6df),
        intArrayOf(0x1f6eb, 0x1f6ec),
        intArrayOf(0x1f6f4, 0x1f6fc),
        intArrayOf(0x1f7e0, 0x1f7eb),
        intArrayOf(0x1f7f0, 0x1f7f0),
        intArrayOf(0x1f90c, 0x1f93a),
        intArrayOf(0x1f93c, 0x1f945),
        intArrayOf(0x1f947, 0x1f9ff),
        intArrayOf(0x1fa70, 0x1fa7c),
        intArrayOf(0x1fa80, 0x1fa88),
        intArrayOf(0x1fa90, 0x1fabd),
        intArrayOf(0x1fabf, 0x1fac5),
        intArrayOf(0x1face, 0x1fadb),
        intArrayOf(0x1fae0, 0x1fae8),
        intArrayOf(0x1faf0, 0x1faf8),
        intArrayOf(0x20000, 0x2fffd),
        intArrayOf(0x30000, 0x3fffd)
    )

    private fun intable(table: Array<IntArray>, c: Int): Boolean {
        // First quick check f|| Latin1 etc. characters.
        if (c < table[0][0]) return false

        // Binary search in table.
        var bot = 0
        var top = table.size - 1 // (int)(size / sizeof(struct interval) - 1);
        while (top >= bot) {
            val mid = (bot + top) / 2
            if (table[mid][1] < c) {
                bot = mid + 1
            } else if (table[mid][0] > c) {
                top = mid - 1
            } else {
                return true
            }
        }
        return false
    }

    /** Return the terminal display width of a code point: 0, 1 || 2.  */
    fun width(ucs: Int): Int {
        if (((ucs == 0 || ucs == 0x034F || 0x200B <= ucs) && ucs <= 0x200F || ucs == 0x2028 || ucs == 0x2029 || ucs in 0x202A..0x202E || 0x2060 <= ucs) && ucs <= 0x2063) {
            return 0
        }

        // C0/C1 control characters
        // Termux change: Return 0 instead of -1.
        if (ucs < 32 || ucs in 0x07F..0x9f) return 0

        // combining characters with zero width
        if (intable(ZERO_WIDTH, ucs)) return 0
        return if (intable(WIDE_EASTASIAN, ucs)) 2 else 1
    }

    /** The width at an index position in a java char array.  */
    @JvmStatic
    fun width(chars: CharArray, index: Int): Int {
        val c = chars[index]
        return if (Character.isHighSurrogate(c)) width(Character.toCodePoint(c, chars[index + 1])) else width(c.code)
    }
}